@@ -21,7 +21,6 @@ from collections import OrderedDict
 import datetime
 from functools import partial
 import os.path
-import sys
 
 from PyQt5 import (
     QtCore,
@@ -30,12 +29,14 @@ from PyQt5 import (
 )
 
 from picard import (
+    PICARD_APP_ID,
     config,
     log,
 )
 from picard.album import Album
 from picard.cluster import Cluster
 from picard.const import PROGRAM_UPDATE_LEVELS
+from picard.const.sys import IS_MACOS
 from picard.file import File
 from picard.formats import supported_formats
 from picard.plugin import ExtensionPoint
@@ -73,6 +74,7 @@ from picard.ui.passworddialog import (
     PasswordDialog,
     ProxyDialog,
 )
+from picard.ui.playertoolbar import Player
 from picard.ui.searchdialog.album import AlbumSearchDialog
 from picard.ui.searchdialog.track import TrackSearchDialog
 from picard.ui.tagsfromfilenames import TagsFromFileNamesDialog
@@ -81,7 +83,8 @@ from picard.ui.util import (
     find_starting_directory,
 )
 
-ui_init = ExtensionPoint()
+
+ui_init = ExtensionPoint(label='ui_init')
 
 
 def register_ui_init(function):
@@ -102,6 +105,8 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
         config.BoolOption("persist", "view_toolbar", True),
         config.BoolOption("persist", "view_file_browser", False),
         config.TextOption("persist", "current_directory", ""),
+        config.FloatOption("persist", "mediaplayer_playback_rate", 1.0),
+        config.IntOption("persist", "mediaplayer_volume", 50),
     ]
 
     def __init__(self, parent=None):
@@ -109,27 +114,32 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
         self.selected_objects = []
         self.ignore_selection_changes = False
         self.toolbar = None
+        self.player = None
+        player = Player(self)
+        if player.available:
+            self.player = player
+            self.player.error.connect(self._on_player_error)
         self.setupUi()
 
     def setupUi(self):
         self.setWindowTitle(_("MusicBrainz Picard"))
         icon = QtGui.QIcon()
-        icon.addFile(":/images/16x16/org.musicbrainz.Picard.png", QtCore.QSize(16, 16))
-        icon.addFile(":/images/24x24/org.musicbrainz.Picard.png", QtCore.QSize(24, 24))
-        icon.addFile(":/images/32x32/org.musicbrainz.Picard.png", QtCore.QSize(32, 32))
-        icon.addFile(":/images/48x48/org.musicbrainz.Picard.png", QtCore.QSize(48, 48))
-        icon.addFile(":/images/128x128/org.musicbrainz.Picard.png", QtCore.QSize(128, 128))
-        icon.addFile(":/images/256x256/org.musicbrainz.Picard.png", QtCore.QSize(256, 256))
+        for size in (16, 24, 32, 48, 128, 256):
+            icon.addFile(
+                ":/images/{size}x{size}/{app_id}.png".format(
+                    size=size, app_id=PICARD_APP_ID),
+                QtCore.QSize(size, size)
+            )
         self.setWindowIcon(icon)
 
-        self.show_close_window = sys.platform == "darwin"
+        self.show_close_window = IS_MACOS
 
         self.create_actions()
         self.create_statusbar()
         self.create_toolbar()
         self.create_menus()
 
-        if sys.platform == "darwin":
+        if IS_MACOS:
             self.setUnifiedTitleAndToolBarOnMac(True)
             self.toolbar.setMovable(False)
             self.search_toolbar.setMovable(False)
@@ -152,6 +162,7 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
 
         self.logDialog = LogView(self)
         self.historyDialog = HistoryView(self)
+        self.optionsDialog = None
 
         bottomLayout = QtWidgets.QHBoxLayout()
         bottomLayout.setContentsMargins(0, 0, 0, 0)
@@ -174,7 +185,7 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
     def keyPressEvent(self, event):
         # On macOS Command+Backspace triggers the so called "Forward Delete".
         # It should be treated the same as the Delete button.
-        is_forward_delete = sys.platform == 'darwin' and \
+        is_forward_delete = IS_MACOS and \
             event.key() == QtCore.Qt.Key_Backspace and \
             event.modifiers() & QtCore.Qt.ControlModifier
         if event.matches(QtGui.QKeySequence.Delete) or is_forward_delete:
@@ -196,6 +207,9 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
         if config.setting["quit_confirmation"] and not self.show_quit_confirmation():
             event.ignore()
             return
+        if self.player:
+            config.persist['mediaplayer_playback_rate'] = self.player.playback_rate()
+            config.persist['mediaplayer_volume'] = self.player.volume()
         self.saveWindowState()
         event.accept()
 
@@ -462,18 +476,7 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
         if discid is None:
             log.warning("CDROM: discid library not found - Lookup CD functionality disabled")
         else:
-            drives = get_cdrom_drives()
-            if not drives:
-                log.warning("CDROM: No CD-ROM drives found - Lookup CD functionality disabled")
-            else:
-                shortcut_drive = config.setting["cd_lookup_device"].split(",")[0] if len(drives) > 1 else ""
-                self.cd_lookup_action.setEnabled(True)
-                for drive in drives:
-                    action = self.cd_lookup_menu.addAction(drive)
-                    if drive == shortcut_drive:
-                        # Clear existing shortcode on main action and assign it to sub-action
-                        self.cd_lookup_action.setShortcut(QtGui.QKeySequence())
-                        action.setShortcut(QtGui.QKeySequence(_("Ctrl+K")))
+            thread.run_task(get_cdrom_drives, self._update_cd_lookup_actions)
 
         self.analyze_action = QtWidgets.QAction(icontheme.lookup('picard-analyze'), _("&Scan"), self)
         self.analyze_action.setStatusTip(_("Use AcoustID audio fingerprint to identify the files by the actual music, even if they have no metadata"))
@@ -561,6 +564,37 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
             self.check_update_action.setMenuRole(QtWidgets.QAction.ApplicationSpecificRole)
             self.check_update_action.triggered.connect(self.do_update_check)
 
+    def _update_cd_lookup_actions(self, result=None, error=None):
+        if error:
+            log.error("CDROM: Error on CD-ROM drive detection: %r", error)
+        else:
+            self.update_cd_lookup_drives(result)
+
+    def update_cd_lookup_drives(self, drives):
+        if not drives:
+            log.warning("CDROM: No CD-ROM drives found - Lookup CD functionality disabled")
+        else:
+            shortcut_drive = config.setting["cd_lookup_device"].split(",")[0] if len(drives) > 1 else ""
+            self.cd_lookup_action.setEnabled(discid is not None)
+            self.cd_lookup_menu.clear()
+            for drive in drives:
+                action = self.cd_lookup_menu.addAction(drive)
+                action.setData(drive)
+                if drive == shortcut_drive:
+                    # Clear existing shortcode on main action and assign it to sub-action
+                    self.cd_lookup_action.setShortcut(QtGui.QKeySequence())
+                    action.setShortcut(QtGui.QKeySequence(_("Ctrl+K")))
+        self._update_cd_lookup_button()
+
+    def _update_cd_lookup_button(self):
+        if len(self.cd_lookup_menu.actions()) > 1:
+            button = self.toolbar.widgetForAction(self.cd_lookup_action)
+            if button:
+                button.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)
+            self.cd_lookup_action.setMenu(self.cd_lookup_menu)
+        else:
+            self.cd_lookup_action.setMenu(None)
+
     def toggle_rename_files(self, checked):
         config.setting["rename_files"] = checked
 
@@ -611,6 +645,8 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
         menu.addSeparator()
         menu.addAction(self.show_toolbar_action)
         menu.addAction(self.search_toolbar_toggle_action)
+        if self.player:
+            menu.addAction(self.player_toolbar_toggle_action)
         menu = self.menuBar().addMenu(_("&Options"))
         menu.addAction(self.enable_renaming_action)
         menu.addAction(self.enable_moving_action)
@@ -619,10 +655,7 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
         menu.addAction(self.options_action)
         menu = self.menuBar().addMenu(_("&Tools"))
         menu.addAction(self.refresh_action)
-        if len(self.cd_lookup_menu.actions()) > 1:
-            menu.addMenu(self.cd_lookup_menu)
-        else:
-            menu.addAction(self.cd_lookup_action)
+        menu.addAction(self.cd_lookup_action)
         menu.addAction(self.autotag_action)
         menu.addAction(self.analyze_action)
         menu.addAction(self.cluster_action)
@@ -649,11 +682,17 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
     def update_toolbar_style(self):
         if config.setting["toolbar_show_labels"]:
             self.toolbar.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
+            if self.player:
+                self.player.toolbar.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
         else:
             self.toolbar.setToolButtonStyle(QtCore.Qt.ToolButtonIconOnly)
+            if self.player:
+                self.player.toolbar.setToolButtonStyle(QtCore.Qt.ToolButtonIconOnly)
 
     def create_toolbar(self):
         self.create_search_toolbar()
+        if self.player:
+            self.create_player_toolbar()
         self.create_action_toolbar()
 
     def create_action_toolbar(self):
@@ -674,10 +713,7 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
         for action in config.setting['toolbar_layout']:
             if action == 'cd_lookup_action':
                 add_toolbar_action(self.cd_lookup_action)
-                if len(self.cd_lookup_menu.actions()) > 1:
-                    button = toolbar.widgetForAction(self.cd_lookup_action)
-                    button.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)
-                    button.setMenu(self.cd_lookup_menu)
+                self._update_cd_lookup_button()
             elif action == 'separator':
                 toolbar.addSeparator()
             else:
@@ -687,6 +723,13 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
                     log.warning('Warning: Unknown action name "%r" found in config. Ignored.', action)
         self.show_toolbar()
 
+    def create_player_toolbar(self):
+        """"Create a toolbar with internal player control elements"""
+        toolbar = self.player.create_toolbar()
+        self.addToolBar(QtCore.Qt.BottomToolBarArea, toolbar)
+        self.player_toolbar_toggle_action = toolbar.toggleViewAction()
+        toolbar.hide()  # Hide by default
+
     def create_search_toolbar(self):
         self.search_toolbar = toolbar = self.addToolBar(_("Search"))
         self.search_toolbar_toggle_action = self.search_toolbar.toggleViewAction()
@@ -848,8 +891,15 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
         self.show_options("about")
 
     def show_options(self, page=None):
-        dialog = OptionsDialog(page, self)
-        dialog.exec_()
+        if not self.optionsDialog:
+            self.optionsDialog = OptionsDialog(page, self)
+            self.optionsDialog.finished.connect(self.on_options_closed)
+        self.optionsDialog.show()
+        self.optionsDialog.raise_()
+        self.optionsDialog.activateWindow()
+
+    def on_options_closed(self):
+        self.optionsDialog = None
 
     def show_help(self):
         webbrowser2.goto('documentation')
@@ -897,6 +947,9 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
         for file in files:
             QtGui.QDesktopServices.openUrl(self._openUrl(file.filename))
 
+    def _on_player_error(self, error, msg):
+        self.set_statusbar_message(msg, echo=log.warning, translate=None)
+
     def open_folder(self):
         files = self.tagger.get_files_from_objects(self.selected_objects)
         folders = set([os.path.dirname(f.filename) for f in files])
@@ -1011,6 +1064,9 @@ class MainWindow(QtWidgets.QMainWindow, PreserveGeometry):
         # Clear any existing status bar messages
         self.set_statusbar_message("")
 
+        if self.player:
+            self.player.set_objects(self.selected_objects)
+
         if len(objects) == 1:
             obj = list(objects)[0]
             if isinstance(obj, File):
